/* slurp-resource.c
 *
 * Copyright © 2017 Julian Richen <julian@richen.io>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define G_LOG_DOMAIN "slurp-resource"

#include "slurp-resource.h"

static gchar *
supported_schemes[] = {
    "http",
    "https",
    NULL
};

typedef struct
{
    const gchar *default_target;
    GHashTable *schemes;

    gchar *source;
    gchar *target;
    gboolean overwrite;
    GFile *file;
    guint64 total;
    guint64 downloaded;
    guint64 updated;
} SlurpResourcePrivate;

G_DEFINE_TYPE_WITH_PRIVATE (SlurpResource, slurp_resource, G_TYPE_OBJECT)

enum {
    PROP_0,
    PROP_SOURCE,
    PROP_TARGET,
    PROP_OVERWRITE,
    PROP_FILE,
    PROP_TOTAL,
    PROP_DOWNLOADED,
    PROP_UPDATED,
    N_PROPS
};

static GParamSpec *properties[N_PROPS];

enum {
    STARTED,
    FINISHED,
    CANCELLED,
    PROGRESS,
    LAST_SIGNAL
};

static guint signals[LAST_SIGNAL];

enum {
    SLURP_RESOURCE_ERROR,
    SLURP_RESOURCE_ERROR_SOURCE,
    SLURP_RESOURCE_ERROR_SOURCE_MISSING,
    SLURP_RESOURCE_ERROR_SOURCE_INVALID_URI,
    SLURP_RESOURCE_ERROR_SOURCE_INVALID_SCHEME,
    SLURP_RESOURCE_ERROR_TARGET,
    SLURP_RESOURCE_ERROR_TARGET_MISSING,
    SLURP_RESOURCE_ERROR_TARGET_INVALID_URI,
    SLURP_RESOURCE_ERROR_TARGET_INVALID_SCHEME,
    SLURP_RESOURCE_ERROR_WRITABLE
};

/**
 * slurp_resource_is_support_scheme:
 * @scheme: a string representing the scheme
 *
 * Checks whether a scheme is supported by slurp. This function does no
 * blocking IO.
 *
 * Returns: %TRUE if scheme is supported, or %FALSE if not supported
 */
gboolean
slurp_resource_is_support_scheme (SlurpResource *self,
                                  const gchar   *scheme)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    return g_hash_table_contains (priv->schemes, scheme);
}

gboolean
slurp_resource_is_valid_source (SlurpResource  *self,
                                const gchar    *uri,
                                GCancellable   *cancellable,
                                GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    GFile *file;

    if (g_strcmp0 (uri, "") == 0)
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_SOURCE_INVALID_SCHEME,
                     "Source can not be empty");
        goto failed;
    }

    // Check URI

    file = g_file_new_for_uri (uri);

    if (!slurp_resource_is_support_scheme (self, g_file_get_uri_scheme (file)))
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_SOURCE_INVALID_SCHEME,
                     "Source is not a support scheme type.");
    }

    g_object_unref (file);
    return TRUE;

failed:
    return FALSE;
}

gboolean
slurp_resource_is_valid_target (SlurpResource  *self,
                                const gchar    *uri,
                                GCancellable   *cancellable,
                                GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (!priv->file)
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_TARGET_MISSING,
                     "Target is not set");
        return FALSE;
    }

    return TRUE;
}

gboolean
slurp_resource_is_writable (SlurpResource  *self,
                            const gchar    *uri,
                            GCancellable   *cancellable,
                            GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (!priv->file)
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_TARGET_MISSING,
                     "Target is not set");
        return FALSE;
    }

    if (!priv->overwrite && g_file_query_exists (priv->file, cancellable))
    {
        g_set_error (error,
                     SLURP_RESOURCE_ERROR,
                     SLURP_RESOURCE_ERROR_WRITABLE,
                     "Target is not writable");
        return FALSE;
    }

    return TRUE;
}

gboolean
slurp_resource_is_valid (SlurpResource  *self,
                         GCancellable   *cancellable,
                         GError        **error)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), FALSE);
    g_return_val_if_fail (cancellable == NULL || G_IS_CANCELLABLE (cancellable), FALSE);

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    g_debug ("Source: %s", priv->source);
    g_debug ("Target: %s", priv->target);

    return TRUE;

/*
    GError *internal_error = NULL;

    if (!slurp_resource_is_valid_source (self, cancellable, &internal_error))
    {
        g_propagate_error (error, internal_error);
        goto out;
    }

    if (!slurp_resource_is_valid_target (self, cancellable, &internal_error))
    {
        g_propagate_error (error, internal_error);
        goto out;
    }

    if (!slurp_resource_is_writable (self, cancellable, &internal_error))
    {
        g_propagate_error (error, internal_error);
        goto out;
    }

    return TRUE;

out:
    g_clear_error (&internal_error);
    return FALSE;
*/
}

static void
_resource_generate_target (SlurpResource *self)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));

    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    if (priv->source == NULL)
        return;

    if (priv->target == NULL || g_strcmp0 (priv->target, priv->default_target) == 0)
    {
        priv->target =
            g_build_filename (priv->default_target,
                              g_path_get_basename (priv->source),
                              NULL);
    }

    if (slurp_utils_uri_is_directory (priv->target))
    {
        priv->target =
            g_build_filename (priv->target,
                              g_path_get_basename (priv->source),
                              NULL);
    }
}

static void
_resource_updated (SlurpResource *self)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));

    GValue value = G_VALUE_INIT;

    g_value_init (&value, G_TYPE_UINT64);
    g_value_set_uint64 (&value, g_get_monotonic_time());
    g_object_set_property (G_OBJECT (self), "updated", &value);
    g_value_unset (&value);
}

static void
slurp_resource_finalize (GObject *object)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    g_clear_pointer (&priv->default_target, g_free);
    g_clear_pointer (&priv->schemes, g_object_unref);

    g_clear_pointer (&priv->source, g_free);
    g_clear_pointer (&priv->target, g_free);
    g_clear_pointer (&priv->file, g_object_unref);

    G_OBJECT_CLASS (slurp_resource_parent_class)->finalize (object);
}

static void
slurp_resource_get_property (GObject    *object,
                             guint       prop_id,
                             GValue     *value,
                             GParamSpec *pspec)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SOURCE:
            g_value_set_string (value, priv->source);
            break;

        case PROP_TARGET:
            g_value_set_string (value, priv->target);
            break;

        case PROP_OVERWRITE:
            g_value_set_boolean (value, priv->overwrite);
            break;

        case PROP_FILE:
            g_value_set_object (value, priv->file);
            break;

        case PROP_TOTAL:
            g_value_set_uint (value, priv->total);
            break;

        case PROP_DOWNLOADED:
            g_value_set_uint (value, priv->downloaded);
            break;

        case PROP_UPDATED:
            g_value_set_uint (value, priv->updated);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_resource_set_property (GObject      *object,
                             guint         prop_id,
                             const GValue *value,
                             GParamSpec   *pspec)
{
    SlurpResource *self = SLURP_RESOURCE (object);
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);

    switch (prop_id)
    {
        case PROP_SOURCE:
            priv->source = g_value_dup_string (value);
            _resource_generate_target (self);
            break;

        case PROP_TARGET:
            priv->target = g_value_dup_string (value);
            _resource_generate_target (self);
            break;

        case PROP_OVERWRITE:
            priv->overwrite = g_value_get_boolean (value);
            break;

        case PROP_FILE:
            priv->file = g_value_dup_object (value);
            break;

        case PROP_TOTAL:
            priv->total = g_value_get_uint (value);
            _resource_updated (self);
            break;

        case PROP_DOWNLOADED:
            priv->downloaded = g_value_get_uint (value);
            _resource_updated (self);
            break;

        case PROP_UPDATED:
            priv->updated = g_value_get_uint (value);
            g_signal_emit (G_OBJECT (self), signals[PROGRESS], 0, priv->downloaded, priv->total);
            break;

        default:
            G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
    }
}

static void
slurp_resource_class_init (SlurpResourceClass *klass)
{
    GObjectClass *object_class = G_OBJECT_CLASS (klass);

    object_class->finalize = slurp_resource_finalize;
    object_class->get_property = slurp_resource_get_property;
    object_class->set_property = slurp_resource_set_property;

	/**
	 * SlurpResource:source:
	 *
	 * The sourceal resource to download from in the form of a URI.
	 */
    properties[PROP_SOURCE] =
        g_param_spec_string ("source",
                             "Sourceal resource URI",
                             "The sourceal resource to download from in the form of a URI.",
                             NULL,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT);

	/**
	 * SlurpResource:target:
	 *
	 * The target which the resource will be saved to in the form of a URI
	 */
    properties[PROP_TARGET] =
        g_param_spec_string ("target",
                             "Target resource URI",
                             "The target which the resource will be saved to in the form of a URI.",
                             // TODO: replace with priv->default_target
                             g_get_user_special_dir (G_USER_DIRECTORY_DOWNLOAD),
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT);

	/**
	 * SlurpResource:overwrite:
	 *
	 * Should files be overwritten by default.
     *
     * Recommended: %FALSE
	 */
    properties[PROP_OVERWRITE] =
        g_param_spec_boolean ("overwrite",
                              "Overwrite files",
                              "Should files be overwritten by default? Default: FALSE",
                              FALSE,
                              G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT);

	/**
	 * SlurpResource:file:
	 *
	 * The #SlurpResource:target: file being written to as a #GFile.
	 */
    properties[PROP_FILE] =
        g_param_spec_object ("file",
                             "a GFile",
                             "A GFile used to write the resource to the system.",
                             G_TYPE_FILE,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE);

	/**
	 * SlurpResource:total:
	 *
	 * The total size o
	 */
    properties[PROP_TOTAL] =
        g_param_spec_uint64 ("total",
                             "Total bytes",
                             "The total number of bytes of the resource.",
                             -1,
                             G_MAXUINT64,
                             0,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT);

	/**
	 * SlurpResource::
	 *
	 *
	 */
    properties[PROP_DOWNLOADED] =
        g_param_spec_uint64 ("downloaded",
                             "Total downloaded bytes",
                             "The total number of downloaded byte of the resource.",
                             0,
                             G_MAXUINT64,
                             0,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT);

	/**
	 * SlurpResource::
	 *
	 *
	 */
    properties[PROP_UPDATED] =
        g_param_spec_uint64 ("updated",
                             "Last time progress was updated",
                             "The last time the progress of the resource download was updated using the system monotonic time.",
                             0,
                             G_MAXUINT64,
                             0,
                             G_PARAM_STATIC_STRINGS | G_PARAM_READWRITE | G_PARAM_CONSTRUCT);

    g_object_class_install_properties (object_class,
                                       N_PROPS,
                                       properties);

    signals[STARTED] =
        g_signal_new ("started",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);

    signals[FINISHED] =
        g_signal_new ("finished",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);
    signals[CANCELLED] =
        g_signal_new ("cancelled",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      0,
                      NULL);

    signals[PROGRESS] =
        g_signal_new ("progress",
                      G_TYPE_FROM_CLASS (object_class),
                      G_SIGNAL_RUN_LAST,
                      0,
                      NULL,
                      NULL,
                      NULL,
                      G_TYPE_NONE,
                      2,
                      G_TYPE_UINT64,
                      G_TYPE_UINT64);
}

static void
slurp_resource_init (SlurpResource *self)
{
    SlurpResourcePrivate *priv = slurp_resource_get_instance_private (self);
    gint i;

    priv->default_target = g_get_user_special_dir (G_USER_DIRECTORY_DOWNLOAD);
    priv->schemes = g_hash_table_new (g_str_hash, g_str_equal);

    for (i = 0; supported_schemes[i] != NULL; ++i)
    {
        g_hash_table_add (priv->schemes, supported_schemes[i]);
    }
}

/*
 * *****************************************************************************
 * API
 * *****************************************************************************
 */

SlurpResource *
slurp_resource_new (void)
{
    return g_object_new (SLURP_TYPE_RESOURCE, NULL);
}

gchar *
slurp_resource_get_source (SlurpResource *self)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), NULL);

    GValue value = G_VALUE_INIT;
    gchar *retval;

    g_value_init (&value, G_TYPE_STRING);
    g_object_get_property (G_OBJECT (self), "source", &value);
    retval = g_strdup (g_value_get_string (&value));
    g_value_unset (&value);
    return retval;
}

void
slurp_resource_set_source (SlurpResource *self,
                           const gchar   *source)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));

    GValue value = G_VALUE_INIT;

    g_value_init (&value, G_TYPE_STRING);
    g_value_set_string (&value, source);
    g_object_set_property (G_OBJECT (self), "source", &value);
    g_value_unset (&value);
}

gchar *
slurp_resource_get_target (SlurpResource *self)
{
    g_return_val_if_fail (SLURP_IS_RESOURCE (self), NULL);

    GValue value = G_VALUE_INIT;
    gchar *retval;

    g_value_init (&value, G_TYPE_STRING);
    g_object_get_property (G_OBJECT (self), "target", &value);
    retval = g_strdup (g_value_get_string (&value));
    g_value_unset (&value);
    return retval;
}

void
slurp_resource_set_target (SlurpResource *self,
                           const gchar   *target)
{
    g_return_if_fail (SLURP_IS_RESOURCE (self));

    GValue value = G_VALUE_INIT;

    g_value_init (&value, G_TYPE_STRING);
    g_value_set_string (&value, target);
    g_object_set_property (G_OBJECT (self), "target", &value);
    g_value_unset (&value);
}
